@@ -1,4 +1,4 @@
-//  Copyright (c) 2022-2024 Fredrik Mellbin
+//  Copyright (c) 2022-2025 Fredrik Mellbin
 //
 //  Permission is hereby granted, free of charge, to any person obtaining a copy
 //  of this software and associated documentation files (the "Software"), to deal
@@ -50,9 +50,9 @@ static void BSInit() {
 struct BestVideoSourceData {
     VSVideoInfo VI = {};
     std::unique_ptr<BestVideoSource> V;
-    int64_t FPSNum;
-    int64_t FPSDen;
-    bool RFF;
+    int64_t FPSNum = -1;
+    int64_t FPSDen = -1;
+    bool RFF = false;
 };
 
 static const VSFrame *VS_CC BestVideoSourceGetFrame(int n, int ActivationReason, void *InstanceData, void **, VSFrameContext *FrameCtx, VSCore *Core, const VSAPI *vsapi) {
@@ -108,12 +108,11 @@ static const VSFrame *VS_CC BestVideoSourceGetFrame(int n, int ActivationReason,
             return nullptr;
         }
 
-        const VideoProperties &VP = D->V->GetVideoProperties();
         VSMap *Props = vsapi->getFramePropertiesRW(Dst);
         if (AlphaDst)
             vsapi->mapConsumeFrame(Props, "_Alpha", AlphaDst, maAppend);
 
-        SetSynthFrameProperties(Src, VP, D->RFF, D->V->GetFrameIsTFF(n, D->RFF),
+        SetSynthFrameProperties(n, Src, *D->V, D->RFF, D->V->GetFrameIsTFF(n, D->RFF),
             [Props, vsapi](const char *Name, int64_t V) { vsapi->mapSetInt(Props, Name, V, maAppend); },
             [Props, vsapi](const char *Name, double V) { vsapi->mapSetFloat(Props, Name, V, maAppend); },
             [Props, vsapi](const char *Name, const char *V, int Size, bool Utf8) { vsapi->mapSetData(Props, Name, V, Size, Utf8 ? dtUtf8 : dtBinary, maAppend); });
@@ -139,18 +138,27 @@ static void VS_CC CreateBestVideoSource(const VSMap *In, VSMap *Out, void *, VSC
     int Track = vsapi->mapGetIntSaturated(In, "track", 0, &err);
     if (err)
         Track = -1;
-    bool VariableFormat = !!vsapi->mapGetInt(In, "variableformat", 0, &err);
+    int ViewID = vsapi->mapGetIntSaturated(In, "viewid", 0, &err);
+    int VariableFormat = vsapi->mapGetIntSaturated(In, "variableformat", 0, &err);
+    if (err)
+        VariableFormat = -1;
     int Threads = vsapi->mapGetIntSaturated(In, "threads", 0, &err);
     int StartNumber = vsapi->mapGetIntSaturated(In, "start_number", 0, &err);
     if (err)
         StartNumber = -1;
     bool ShowProgress = !!vsapi->mapGetInt(In, "showprogress", 0, &err);
+    if (err)
+        ShowProgress = true;
     int ExtraHWFrames = vsapi->mapGetIntSaturated(In, "extrahwframes", 0, &err);
     if (err)
         ExtraHWFrames = 9;
     int CacheMode = vsapi->mapGetIntSaturated(In, "cachemode", 0, &err);
     if (err)
         CacheMode = 1;
+    int MaxDecoders = vsapi->mapGetIntSaturated(In, "maxdecoders", 0, &err);
+    bool HWFallback = !!vsapi->mapGetInt(In, "hwfallback", 0, &err);
+    if (err)
+        HWFallback = true;
 
     std::map<std::string, std::string> Opts;
     if (vsapi->mapGetInt(In, "enable_drefs", 0, &err))
@@ -180,34 +188,58 @@ static void VS_CC CreateBestVideoSource(const VSMap *In, VSMap *Out, void *, VSC
         if (ShowProgress) {
             auto NextUpdate = std::chrono::high_resolution_clock::now();
             int LastValue = -1;
-            D->V.reset(new BestVideoSource(Source, HWDevice ? HWDevice : "", ExtraHWFrames, Track, VariableFormat, Threads, CacheMode, CachePath ? CachePath : "", &Opts,
-                [vsapi, Core, &NextUpdate, &LastValue](int Track, int64_t Cur, int64_t Total) {
-                    if (NextUpdate < std::chrono::high_resolution_clock::now()) {
-                        if (Total == INT64_MAX && Cur == Total) {
-                            vsapi->logMessage(mtInformation, ("VideoSource track #" + std::to_string(Track) + " indexing complete").c_str(), Core);
-                        } else {
-                            int PValue = (Total > 0) ? static_cast<int>((static_cast<double>(Cur) / static_cast<double>(Total)) * 100) : static_cast<int>(Cur / (1024 * 1024));
-                            if (PValue != LastValue) {
-                                vsapi->logMessage(mtInformation, ("VideoSource track #" + std::to_string(Track) + " index progress " + std::to_string(PValue) + ((Total > 0) ? "%" : "MB")).c_str(), Core);
-                                LastValue = PValue;
-                                NextUpdate = std::chrono::high_resolution_clock::now() + std::chrono::seconds(1);
-                            }
+            auto ProgressCB = [vsapi, Core, &NextUpdate, &LastValue](int Track, int64_t Cur, int64_t Total) {
+                if (NextUpdate < std::chrono::high_resolution_clock::now()) {
+                    if (Total == INT64_MAX && Cur == Total) {
+                        vsapi->logMessage(mtInformation, ("VideoSource track #" + std::to_string(Track) + " indexing complete").c_str(), Core);
+                    } else {
+                        int PValue = (Total > 0) ? static_cast<int>((static_cast<double>(Cur) / static_cast<double>(Total)) * 100) : static_cast<int>(Cur / (1024 * 1024));
+                        if (PValue != LastValue) {
+                            vsapi->logMessage(mtInformation, ("VideoSource track #" + std::to_string(Track) + " index progress " + std::to_string(PValue) + ((Total > 0) ? "%" : "MB")).c_str(), Core);
+                            LastValue = PValue;
+                            NextUpdate = std::chrono::high_resolution_clock::now() + std::chrono::seconds(1);
                         }
                     }
-                    return true;
-                }));
-
+                }
+                return true;
+                };
+            try {
+                D->V.reset(new BestVideoSource(Source, HWDevice ? HWDevice : "", ExtraHWFrames, Track, ViewID, Threads, CacheMode, CachePath ? CachePath : "", &Opts, ProgressCB));
+            } catch (BestSourceHWDecoderException &) {
+                if (HWFallback) {
+                    D->V.reset(new BestVideoSource(Source, "", ExtraHWFrames, Track, ViewID, Threads, CacheMode, CachePath ? CachePath : "", &Opts, ProgressCB));
+                } else {
+                    vsapi->logMessage(mtInformation, ("VideoSource track #" + std::to_string(Track) + " using CPU decoding fallback").c_str(), Core);
+                    throw;
+                }
+            }
         } else {
-            D->V.reset(new BestVideoSource(Source, HWDevice ? HWDevice : "", ExtraHWFrames, Track, VariableFormat, Threads, CacheMode, CachePath ? CachePath : "", &Opts));
+            try {
+                D->V.reset(new BestVideoSource(Source, HWDevice ? HWDevice : "", ExtraHWFrames, Track, ViewID, Threads, CacheMode, CachePath ? CachePath : "", &Opts));
+            } catch (BestSourceHWDecoderException &) {
+                if (HWFallback) {
+                    D->V.reset(new BestVideoSource(Source, "", ExtraHWFrames, Track, ViewID, Threads, CacheMode, CachePath ? CachePath : "", &Opts));
+                } else {
+                    vsapi->logMessage(mtInformation, ("VideoSource track #" + std::to_string(Track) + " using CPU decoding fallback").c_str(), Core);
+                    throw;
+                }
+            }
         }
 
-        const VideoProperties &VP = D->V->GetVideoProperties();
-        if (VP.VF.ColorFamily == 0 || !vsapi->queryVideoFormat(&D->VI.format, VP.VF.ColorFamily, VP.VF.Float, VP.VF.Bits, VP.VF.SubSamplingW, VP.VF.SubSamplingH, Core))
+        D->V->SetMaxDecoderInstances(MaxDecoders);
+        D->V->SelectFormatSet(VariableFormat);
+
+        const BSVideoProperties &VP = D->V->GetVideoProperties();
+        if (VP.VF.ColorFamily == 4)
+            throw BestSourceException("Unsupported source colorspace (bayer)");
+        if ((VP.VF.ColorFamily == 0 && VariableFormat != -1) || !vsapi->queryVideoFormat(&D->VI.format, VP.VF.ColorFamily, VP.VF.Float, VP.VF.Bits, VP.VF.SubSamplingW, VP.VF.SubSamplingH, Core))
             throw BestSourceException("Unsupported video format from decoder (probably less than 8 bit or palette)");
+
+        if (VP.SSModWidth == 0 || VP.SSModHeight == 0)
+            throw BestSourceException("Rounding dimensions down to nearest subsampling multiple leaves nothing to output");
+
         D->VI.width = VP.SSModWidth;
         D->VI.height = VP.SSModHeight;
-        if (VariableFormat)
-            D->VI = {};
         D->VI.numFrames = vsh::int64ToIntS(VP.NumFrames);
         D->VI.fpsNum = VP.FPS.Num;
         D->VI.fpsDen = VP.FPS.Den;
@@ -215,14 +247,9 @@ static void VS_CC CreateBestVideoSource(const VSMap *In, VSMap *Out, void *, VSC
 
         if (D->FPSNum > 0) {
             vsh::reduceRational(&D->FPSNum, &D->FPSDen);
-            if (VP.FPS.Den != D->FPSDen || VP.FPS.Num != D->FPSNum) {
-                D->VI.fpsDen = D->FPSDen;
-                D->VI.fpsNum = D->FPSNum;
-                D->VI.numFrames = std::max(1, static_cast<int>((VP.Duration * D->VI.fpsNum) / D->VI.fpsDen));
-            } else {
-                D->FPSNum = -1;
-                D->FPSDen = 1;
-            }
+            D->VI.fpsDen = D->FPSDen;
+            D->VI.fpsNum = D->FPSNum;
+            D->VI.numFrames = std::max(1, static_cast<int>((VP.Duration * D->VI.fpsNum) * VP.TimeBase.ToDouble() / D->VI.fpsDen + 0.5));
         } else if (D->RFF) {
             D->VI.numFrames = vsh::int64ToIntS(VP.NumRFFFrames);
         }
@@ -307,6 +334,7 @@ static void VS_CC CreateBestAudioSource(const VSMap *In, VSMap *Out, void *, VSC
     int CacheMode = vsapi->mapGetIntSaturated(In, "cachemode", 0, &err);
     if (err)
         CacheMode = 1;
+    int MaxDecoders = vsapi->mapGetIntSaturated(In, "maxdecoders", 0, &err);
 
     std::map<std::string, std::string> Opts;
     if (vsapi->mapGetInt(In, "enable_drefs", 0, &err))
@@ -322,7 +350,7 @@ static void VS_CC CreateBestAudioSource(const VSMap *In, VSMap *Out, void *, VSC
         if (ShowProgress) {
             auto NextUpdate = std::chrono::high_resolution_clock::now();
             int LastValue = -1;
-            D->A.reset(new BestAudioSource(Source, Track, AdjustDelay, false, Threads, CacheMode, CachePath ? CachePath : "", &Opts, DrcScale,
+            D->A.reset(new BestAudioSource(Source, Track, AdjustDelay, Threads, CacheMode, CachePath ? CachePath : "", &Opts, DrcScale,
                 [vsapi, Core, &NextUpdate, &LastValue](int Track, int64_t Cur, int64_t Total) {
                     if (NextUpdate < std::chrono::high_resolution_clock::now()) {
                         if (Total == INT64_MAX && Cur == Total) {
@@ -340,10 +368,13 @@ static void VS_CC CreateBestAudioSource(const VSMap *In, VSMap *Out, void *, VSC
                 }));
 
         } else {
-            D->A.reset(new BestAudioSource(Source, Track, AdjustDelay, false, Threads, CacheMode, CachePath ? CachePath : "", &Opts, DrcScale));
+            D->A.reset(new BestAudioSource(Source, Track, AdjustDelay, Threads, CacheMode, CachePath ? CachePath : "", &Opts, DrcScale));
         }
 
-        const AudioProperties &AP = D->A->GetAudioProperties();
+        D->A->SetMaxDecoderInstances(MaxDecoders);
+        D->A->SelectFormatSet(0);
+
+        const BSAudioProperties &AP = D->A->GetAudioProperties();
         D->Is8Bit = (AP.AF.Bits <= 8);
         if (!vsapi->queryAudioFormat(&D->AI.format, AP.AF.Float, D->Is8Bit ? 16 : AP.AF.Bits, AP.ChannelLayout, Core))
             throw BestSourceException("Unsupported audio format from decoder (probably 8-bit)");
@@ -393,6 +424,35 @@ static void VS_CC GetTrackInfo(const VSMap *In, VSMap *Out, void *, VSCore *core
     }
 }
 
+static void VS_CC GetMetadata(const VSMap *In, VSMap *Out, void *, VSCore *core, const VSAPI *vsapi) {
+    BSInit();
+
+    int err;
+    std::filesystem::path Source = CreateProbablyUTF8Path(vsapi->mapGetData(In, "source", 0, nullptr));
+
+    int Track = vsapi->mapGetIntSaturated(In, "track", 0, &err);
+    if (err)
+        Track = -1;
+
+    std::map<std::string, std::string> Opts;
+    if (vsapi->mapGetInt(In, "enable_drefs", 0, &err))
+        Opts["enable_drefs"] = "1";
+    if (vsapi->mapGetInt(In, "use_absolute_path", 0, &err))
+        Opts["use_absolute_path"] = "1";
+
+    try {
+        std::unique_ptr<BestTrackList> TrackList(new BestTrackList(Source, &Opts));
+        
+        auto Metadata = (Track < 0) ? TrackList->GetFileMetadata() : TrackList->GetTrackMetadata(Track);
+
+        for (const auto &Iter : Metadata) {
+            vsapi->mapSetData(Out, Iter.first.c_str(), Iter.second.c_str(), -1, dtUtf8, maAppend);
+        }
+    } catch (BestSourceException &e) {
+        vsapi->mapSetError(Out, (std::string("Metadata: ") + e.what()).c_str());
+    }
+}
+
 static void VS_CC SetDebugOutput(const VSMap *In, VSMap *, void *, VSCore *, const VSAPI *vsapi) {
     BSInit();
     SetBSDebugOutput(!!vsapi->mapGetInt(In, "enable", 0, nullptr));
@@ -409,9 +469,10 @@ static void VS_CC SetLogLevel(const VSMap *In, VSMap *Out, void *, VSCore *, con
 
 VS_EXTERNAL_API(void) VapourSynthPluginInit2(VSPlugin *plugin, const VSPLUGINAPI *vspapi) {
     vspapi->configPlugin("com.vapoursynth.bestsource", "bs", "Best Source 2", VS_MAKE_VERSION(BEST_SOURCE_VERSION_MAJOR, BEST_SOURCE_VERSION_MINOR), VS_MAKE_VERSION(VAPOURSYNTH_API_MAJOR, 0), 0, plugin);
-    vspapi->registerFunction("VideoSource", "source:data;track:int:opt;variableformat:int:opt;fpsnum:int:opt;fpsden:int:opt;rff:int:opt;threads:int:opt;seekpreroll:int:opt;enable_drefs:int:opt;use_absolute_path:int:opt;cachemode:int:opt;cachepath:data:opt;cachesize:int:opt;hwdevice:data:opt;extrahwframes:int:opt;timecodes:data:opt;start_number:int:opt;showprogress:int:opt;", "clip:vnode;", CreateBestVideoSource, nullptr, plugin);
-    vspapi->registerFunction("AudioSource", "source:data;track:int:opt;adjustdelay:int:opt;threads:int:opt;enable_drefs:int:opt;use_absolute_path:int:opt;drc_scale:float:opt;cachemode:int:opt;cachepath:data:opt;cachesize:int:opt;showprogress:int:opt;", "clip:anode;", CreateBestAudioSource, nullptr, plugin);
+    vspapi->registerFunction("VideoSource", "source:data;track:int:opt;variableformat:int:opt;fpsnum:int:opt;fpsden:int:opt;rff:int:opt;threads:int:opt;seekpreroll:int:opt;enable_drefs:int:opt;use_absolute_path:int:opt;cachemode:int:opt;cachepath:data:opt;cachesize:int:opt;hwdevice:data:opt;extrahwframes:int:opt;timecodes:data:opt;start_number:int:opt;viewid:int:opt;showprogress:int:opt;maxdecoders:int:opt;hwfallback:int:opt;", "clip:vnode;", CreateBestVideoSource, nullptr, plugin);
+    vspapi->registerFunction("AudioSource", "source:data;track:int:opt;adjustdelay:int:opt;threads:int:opt;enable_drefs:int:opt;use_absolute_path:int:opt;drc_scale:float:opt;cachemode:int:opt;cachepath:data:opt;cachesize:int:opt;showprogress:int:opt;maxdecoders:int:opt;", "clip:anode;", CreateBestAudioSource, nullptr, plugin);
     vspapi->registerFunction("TrackInfo", "source:data;enable_drefs:int:opt;use_absolute_path:int:opt;", "mediatype:int;mediatypestr:data;codec:int;codecstr:data;disposition:int;dispositionstr:data;", GetTrackInfo, nullptr, plugin);
+    vspapi->registerFunction("Metadata", "source:data;track:int:opt;enable_drefs:int:opt;use_absolute_path:int:opt;", "any", GetMetadata, nullptr, plugin);
     vspapi->registerFunction("SetDebugOutput", "enable:int;", "", SetDebugOutput, nullptr, plugin);
     vspapi->registerFunction("SetFFmpegLogLevel", "level:int;", "level:int;", SetLogLevel, nullptr, plugin);
 }
