@@ -1,4 +1,4 @@
-//  Copyright (c) 2022-2024 Fredrik Mellbin
+//  Copyright (c) 2022-2025 Fredrik Mellbin
 //
 //  Permission is hereby granted, free of charge, to any person obtaining a copy
 //  of this software and associated documentation files (the "Software"), to deal
@@ -24,6 +24,7 @@
 #include <thread>
 #include <cassert>
 #include <iterator>
+#include <charconv>
 
 #include "../libp2p/p2p_api.h"
 
@@ -32,6 +33,7 @@
 extern "C" {
 #include <libavformat/avformat.h>
 #include <libavcodec/avcodec.h>
+#include <libavutil/opt.h>
 #include <libavutil/imgutils.h>
 #include <libavutil/stereo3d.h>
 #include <libavutil/display.h>
@@ -50,9 +52,11 @@ static bool HasAlpha(const AVPixFmtDescriptor *Desc) {
 static int GetColorFamily(const AVPixFmtDescriptor *Desc) {
     if (!!(Desc->flags & AV_PIX_FMT_FLAG_PAL))
         return 2;
+    if (!!(Desc->flags & AV_PIX_FMT_FLAG_BAYER))
+        return 4;
     if (Desc->nb_components <= 2)
         return 1;
-    else if (Desc->flags & AV_PIX_FMT_FLAG_RGB)
+    else if (Desc->flags & (AV_PIX_FMT_FLAG_RGB | AV_PIX_FMT_FLAG_XYZ))
         return 2;
     else
         return 3;
@@ -99,7 +103,7 @@ bool LWVideoDecoder::DecodeNextFrame(bool SkipOutput) {
                 }
             }
             return true;
-        } else if (Ret == AVERROR(EAGAIN)) {
+        } else if (Ret == AVERROR(EAGAIN) || Ret == AVERROR_INPUT_CHANGED) { // AVERROR_INPUT_CHANGED is only used for two things inside FFmpeg and the other one can't happen, therefore we don't need to care about whether or not variable format is allowed
             if (ReadPacket()) {
                 avcodec_send_packet(CodecContext, Packet);
                 av_packet_unref(Packet);
@@ -114,7 +118,7 @@ bool LWVideoDecoder::DecodeNextFrame(bool SkipOutput) {
     return false;
 }
 
-void LWVideoDecoder::OpenFile(const std::filesystem::path &SourceFile, const std::string &HWDeviceName, int ExtraHWFrames, int Track, bool VariableFormat, int Threads, const std::map<std::string, std::string> &LAVFOpts) {
+void LWVideoDecoder::OpenFile(const std::filesystem::path &SourceFile, const std::string &HWDeviceName, int ExtraHWFrames, int Track, int ViewID, int Threads, const std::map<std::string, std::string> &LAVFOpts) {
     TrackNumber = Track;
 
     AVHWDeviceType Type = AV_HWDEVICE_TYPE_NONE;
@@ -130,8 +134,10 @@ void LWVideoDecoder::OpenFile(const std::filesystem::path &SourceFile, const std
     for (const auto &Iter : LAVFOpts)
         av_dict_set(&Dict, Iter.first.c_str(), Iter.second.c_str(), 0);
 
-    if (avformat_open_input(&FormatContext, SourceFile.u8string().c_str(), nullptr, &Dict) != 0)
+    if (avformat_open_input(&FormatContext, SourceFile.u8string().c_str(), nullptr, &Dict) != 0) {
+        av_dict_free(&Dict);
         throw BestSourceException("Couldn't open '" + SourceFile.u8string() + "'");
+    }
 
     av_dict_free(&Dict);
 
@@ -160,6 +166,14 @@ void LWVideoDecoder::OpenFile(const std::filesystem::path &SourceFile, const std
     if (FormatContext->streams[TrackNumber]->codecpar->codec_type != AVMEDIA_TYPE_VIDEO)
         throw BestSourceException("Not a video track");
 
+    IsLayered = !!(FormatContext->streams[TrackNumber]->disposition & AV_DISPOSITION_MULTILAYER);
+
+    if (ViewID < 0)
+        throw BestSourceException("Invalid view id specified, negative numbers aren't allowed");
+
+    if (ViewID != 0 && !IsLayered)
+        throw BestSourceException("Non-zero view id specified but there is only a single layer");
+
     for (int i = 0; i < static_cast<int>(FormatContext->nb_streams); i++)
         if (i != TrackNumber)
             FormatContext->streams[i]->discard = AVDISCARD_ALL;
@@ -178,7 +192,7 @@ void LWVideoDecoder::OpenFile(const std::filesystem::path &SourceFile, const std
         for (int i = 0;; i++) {
             const AVCodecHWConfig *Config = avcodec_get_hw_config(Codec, i);
             if (!Config)
-                throw BestSourceException("Decoder " + std::string(Codec->name) + " does not support device type " + av_hwdevice_get_type_name(Type));
+                throw BestSourceHWDecoderException("Decoder " + std::string(Codec->name) + " does not support device type " + av_hwdevice_get_type_name(Type));
             if (Config->methods & AV_CODEC_HW_CONFIG_METHOD_HW_DEVICE_CTX &&
                 Config->device_type == Type) {
                 hw_pix_fmt = Config->pix_fmt;
@@ -212,12 +226,6 @@ void LWVideoDecoder::OpenFile(const std::filesystem::path &SourceFile, const std
     CodecContext->apply_cropping = 1;
     CodecContext->flags |= AV_CODEC_FLAG_UNALIGNED;
 
-    // FIXME, implement for newer ffmpeg versions
-    if (!VariableFormat) {
-        // Probably guard against mid-stream format changes
-        CodecContext->flags |= AV_CODEC_FLAG_DROPCHANGED;
-    }
-
     // Full explanation by more clever person available here: https://github.com/Nevcairiel/LAVFilters/issues/113
     if (CodecContext->codec_id == AV_CODEC_ID_H264 && CodecContext->has_b_frames) {
         CodecContext->has_b_frames = 15; // the maximum possible value for h264
@@ -235,14 +243,46 @@ void LWVideoDecoder::OpenFile(const std::filesystem::path &SourceFile, const std
             throw BestSourceException("Couldn't allocate frame");
     }
 
-    if (avcodec_open2(CodecContext, Codec, nullptr) < 0)
+    AVDictionary *CodecDict = nullptr;
+    if (IsLayered)
+        av_dict_set(&CodecDict, "view_ids", std::to_string(ViewID).c_str(), 0);
+
+    if (avcodec_open2(CodecContext, Codec, &CodecDict) < 0) {
+        av_dict_free(&CodecDict);
         throw BestSourceException("Could not open video codec");
+    }
+
+    av_dict_free(&CodecDict);
+
+    // MV-HEVC views
+    if (IsLayered) {
+        unsigned ArrSize = 0;
+        if (av_opt_get_array_size(CodecContext, "view_ids_available", 0, &ArrSize) >= 0 && ArrSize > 0) {
+            std::vector<int> Tmp1;
+            Tmp1.resize(ArrSize);
+            av_opt_get_array(CodecContext, "view_ids_available", 0, 0, ArrSize, AV_OPT_TYPE_INT, Tmp1.data());
+
+            std::vector<int> Tmp2;
+            if (av_opt_get_array_size(CodecContext, "view_pos_available", 0, &ArrSize) >= 0 && ArrSize > 0) {
+                Tmp2.resize(ArrSize);
+                av_opt_get_array(CodecContext, "view_pos_available", 0, 0, ArrSize, AV_OPT_TYPE_INT, Tmp2.data());
+            }
+
+            for (size_t i = 0; i < Tmp1.size(); i++) {
+                if (i < Tmp2.size())
+                    ViewIDs.push_back({ Tmp1[i], Tmp2[i] });
+                else {
+                    ViewIDs.push_back({ Tmp1[i], -1 });
+                }
+            }
+        }
+    }
 }
 
-LWVideoDecoder::LWVideoDecoder(const std::filesystem::path &SourceFile, const std::string &HWDeviceName, int ExtraHWFrames, int Track, bool VariableFormat, int Threads, const std::map<std::string, std::string> &LAVFOpts) {
+LWVideoDecoder::LWVideoDecoder(const std::filesystem::path &SourceFile, const std::string &HWDeviceName, int ExtraHWFrames, int Track, int ViewID, int Threads, const std::map<std::string, std::string> &LAVFOpts) {
     try {
         Packet = av_packet_alloc();
-        OpenFile(SourceFile, HWDeviceName, ExtraHWFrames, Track, VariableFormat, Threads, LAVFOpts);
+        OpenFile(SourceFile, HWDeviceName, ExtraHWFrames, Track, ViewID, Threads, LAVFOpts);
     } catch (...) {
         Free();
         throw;
@@ -282,21 +322,8 @@ void LWVideoDecoder::SetFrameNumber(int64_t N) {
     CurrentFrame = N;
 }
 
-void LWVideoDecoder::GetVideoProperties(VideoProperties &VP) {
-    assert(CurrentFrame == 0);
+void LWVideoDecoder::GetVideoProperties(LWVideoProperties &VP) {
     VP = {};
-    AVFrame *PropFrame = GetNextFrame();
-    assert(PropFrame);
-    if (!PropFrame)
-        return;
-
-    VP.VF.Set(av_pix_fmt_desc_get(static_cast<AVPixelFormat>(PropFrame->format)));
-    VP.FieldBased = !!(PropFrame->flags & AV_FRAME_FLAG_INTERLACED);
-    VP.TFF = !!(PropFrame->flags & AV_FRAME_FLAG_TOP_FIELD_FIRST);
-    VP.Width = CodecContext->width;
-    VP.Height = CodecContext->height;
-    VP.SSModWidth = VP.Width - (VP.Width % (1 << VP.VF.SubSamplingW));
-    VP.SSModHeight = VP.Height - (VP.Height % (1 << VP.VF.SubSamplingH));
 
     VP.FPS = CodecContext->framerate;
     // Set the framerate from the container if the codec framerate is invalid
@@ -321,9 +348,6 @@ void LWVideoDecoder::GetVideoProperties(VideoProperties &VP) {
         VP.FPS.Num = 30;
     }
 
-    if (PropFrame->pts != AV_NOPTS_VALUE)
-        VP.StartTime = (static_cast<double>(FormatContext->streams[TrackNumber]->time_base.num) * PropFrame->pts) / FormatContext->streams[TrackNumber]->time_base.den;
-
     // Set AR variables
     VP.SAR = CodecContext->sample_aspect_ratio;
 
@@ -331,6 +355,8 @@ void LWVideoDecoder::GetVideoProperties(VideoProperties &VP) {
     if (VP.SAR.Num <= 0 || VP.SAR.Den <= 0)
         VP.SAR = FormatContext->streams[TrackNumber]->sample_aspect_ratio;
 
+    VP.ViewIDs = ViewIDs;
+
     // Set stereoscopic 3d type
     VP.Stereo3DType = AV_STEREO3D_2D;
 
@@ -457,7 +483,7 @@ bool LWVideoDecoder::HasSeeked() const {
     return Seeked;
 }
 
-void VideoFormat::Set(const AVPixFmtDescriptor *Desc) {
+void BSVideoFormat::Set(const AVPixFmtDescriptor *Desc) {
     Alpha = HasAlpha(Desc);
     Float = GetSampleTypeIsFloat(Desc);
     ColorFamily = GetColorFamily(Desc);
@@ -620,7 +646,7 @@ static const std::map<AVPixelFormat, p2p_packing> FormatMap = {
 };
 
 bool BestVideoFrame::ExportAsPlanar(uint8_t *const *const Dsts1, const ptrdiff_t *const Stride, uint8_t *AlphaDst, ptrdiff_t AlphaStride) const {
-    if (VF.ColorFamily == 0)
+    if (VF.ColorFamily == 0 || VF.ColorFamily == 4)
         return false;
 
     uint8_t *Dsts[3] = { Dsts1[0], Dsts1[1], Dsts1[2] };
@@ -780,8 +806,13 @@ static std::array<uint8_t, HashSize> GetHash(const AVFrame *Frame) {
     int NumPlanes = 0;
     int SampleSize[4] = {};
 
+    // A bit weird but for planar formats this will yield the samplesize directly since it's identical to the step
+    // For packed formats like YUY2 and others it will get the average packed unit size per pixel
     for (int i = 0; i < desc->nb_components; i++) {
-        SampleSize[desc->comp[i].plane] = std::max(SampleSize[desc->comp[i].plane], desc->comp[i].step);
+        if (SampleSize[desc->comp[i].plane] == 0)
+            SampleSize[desc->comp[i].plane] = desc->comp[i].step;
+        else
+            SampleSize[desc->comp[i].plane] = std::min(SampleSize[desc->comp[i].plane], desc->comp[i].step);
         NumPlanes = std::max(NumPlanes, desc->comp[i].plane + 1);
     }
 
@@ -867,8 +898,28 @@ BestVideoFrame *BestVideoSource::Cache::GetFrame(int64_t N) {
     return nullptr;
 }
 
-BestVideoSource::BestVideoSource(const std::filesystem::path &SourceFile, const std::string &HWDeviceName, int ExtraHWFrames, int Track, bool VariableFormat, int Threads, int CacheMode, const std::filesystem::path &CachePath, const std::map<std::string, std::string> *LAVFOpts, const ProgressFunction &Progress)
-    : Source(SourceFile), HWDevice(HWDeviceName), ExtraHWFrames(!HWDeviceName.empty() ? ExtraHWFrames : 0), VideoTrack(Track), VariableFormat(VariableFormat), Threads(Threads) {
+bool BestVideoSource::NearestCommonFrameRate(BSRational &FPS) {
+    constexpr std::array FPSList = { 24, 25, 30, 48, 50, 60, 100, 120 };
+    double FPSDouble = FPS.ToDouble();
+
+    for (const auto &Iter : FPSList) {
+        const double delta = (Iter - static_cast<double>(Iter) / 1.001) / 2.0;
+        if (fabs(FPSDouble - Iter) < delta) {
+            FPS.Num = Iter;
+            FPS.Den = 1;
+            return true;
+        } else if ((Iter % 25 != 0) && (fabs(FPSDouble - static_cast<double>(Iter) / 1.001) < delta)) {
+            FPS.Num = Iter * 1000;
+            FPS.Den = 1001;
+            return true;
+        }
+    }
+
+    return false;
+}
+
+BestVideoSource::BestVideoSource(const std::filesystem::path &SourceFile, const std::string &HWDeviceName, int ExtraHWFrames, int Track, int ViewID, int Threads, int CacheMode, const std::filesystem::path &CachePath, const std::map<std::string, std::string> *LAVFOpts, const ProgressFunction &Progress)
+    : Source(SourceFile), HWDevice(HWDeviceName), ExtraHWFrames(!HWDeviceName.empty() ? ExtraHWFrames : 0), VideoTrack(Track), ViewID(ViewID), Threads(Threads) {
     // Only make file path absolute if it exists to pass through special protocol paths
     std::error_code ec;
     if (std::filesystem::exists(SourceFile, ec))
@@ -880,38 +931,110 @@ BestVideoSource::BestVideoSource(const std::filesystem::path &SourceFile, const
     if (ExtraHWFrames < 0)
         throw BestSourceException("ExtraHWFrames must be 0 or greater");
     
-    if (CacheMode < 0 || CacheMode > 2)
-        throw BestSourceException("CacheMode must be between 0 and 2");
+    if (CacheMode < 0 || CacheMode > 4)
+        throw BestSourceException("CacheMode must be between 0 and 4");
 
-    std::unique_ptr<LWVideoDecoder> Decoder(new LWVideoDecoder(Source, HWDevice, ExtraHWFrames, VideoTrack, VariableFormat, Threads, LAVFOptions));
+    if (ViewID < 0)
+        throw BestSourceException("ViewID must be 0 or greater");
+
+    std::unique_ptr<LWVideoDecoder> Decoder(new LWVideoDecoder(Source, HWDevice, ExtraHWFrames, VideoTrack, ViewID, Threads, LAVFOptions));
 
     Decoder->GetVideoProperties(VP);
     VideoTrack = Decoder->GetTrack();
     FileSize = Decoder->GetSourceSize();
 
-    if (CacheMode == bcmDisable || !ReadVideoTrackIndex(CachePath)) {
+    if (CacheMode == bcmDisable || !ReadVideoTrackIndex(IsAbsolutePathCacheMode(CacheMode), CachePath)) {
         if (!IndexTrack(Progress))
             throw BestSourceException("Indexing of '" + Source.u8string() + "' track #" + std::to_string(VideoTrack) + " failed");
 
-        if (CacheMode == bcmAlwaysWrite || (CacheMode == bcmAuto && TrackIndex.Frames.size() >= 100))
-            WriteVideoTrackIndex(CachePath);
+        if (ShouldWriteIndex(CacheMode, TrackIndex.Frames.size())) {
+            if (!WriteVideoTrackIndex(IsAbsolutePathCacheMode(CacheMode), CachePath))
+                throw BestSourceException("Failed to write index to '" + CachePath.u8string() + "' for track #" + std::to_string(VideoTrack));
+        }
     }
 
     if (TrackIndex.Frames[0].RepeatPict < 0)
         throw BestSourceException("Found an unexpected RFF quirk, please submit a bug report and attach the source file");
+    
+    for (const auto &Iter : TrackIndex.Frames) {
+        if (Iter.PTS == AV_NOPTS_VALUE) {
+            CanSeekByTime = false;
+            break;
+        }
+    }
 
-    VP.NumFrames = TrackIndex.Frames.size();
-    VP.Duration = (TrackIndex.Frames.back().PTS - TrackIndex.Frames.front().PTS) + std::max<int64_t>(1, TrackIndex.LastFrameDuration);
+    // Framerate and last frame duration guessing fun
+    const auto OriginalFPS = VP.FPS;
+    std::map<int64_t, size_t> DurationHistogram;
 
-    int64_t NumFields = 0;
+    // This part is to compensate for files that most likely have timestamps attached to the wrong frame due to b-frame reordering and timecode attachment fun in broken tools
+    // Does nothing for most files
+    std::vector<int64_t> SortedPTS;
+    SortedPTS.resize(TrackIndex.Frames.size());
+    for (size_t i = 0; i < TrackIndex.Frames.size(); i++)
+        SortedPTS[i] = TrackIndex.Frames[i].PTS;
+    std::sort(SortedPTS.begin(), SortedPTS.end());
 
-    for (auto &Iter : TrackIndex.Frames)
-        NumFields += Iter.RepeatPict + 2;
+    for (size_t i = 0; i < TrackIndex.Frames.size() - 1; i++)
+        if (SortedPTS[i] == AV_NOPTS_VALUE || SortedPTS[i + 1] == AV_NOPTS_VALUE)
+            ++DurationHistogram[AV_NOPTS_VALUE];
+        else
+            ++DurationHistogram[SortedPTS[i + 1] - SortedPTS[i]];
 
-    VP.NumRFFFrames = (NumFields + 1) / 2;
+    std::pair<int64_t, size_t> MostCommonDuration(1, 1);
+    if (!DurationHistogram.empty())
+        MostCommonDuration = *std::max_element(DurationHistogram.begin(), DurationHistogram.end(), [](const std::pair<int64_t, size_t> &p1, const std::pair<int64_t, size_t> &p2) { return p1.second < p2.second; });
 
-    if (VP.NumFrames == VP.NumRFFFrames)
-        RFFState = RFFStateEnum::Unused;
+    int64_t LastFrameDuration = TrackIndex.LastFrameDuration;
+    if (LastFrameDuration <= 0 && !DurationHistogram.empty() && MostCommonDuration.first > 0)
+        LastFrameDuration = MostCommonDuration.first;
+    LastFrameDuration = std::max<int64_t>(1, LastFrameDuration);
+
+    if (TrackIndex.Frames.front().PTS != AV_NOPTS_VALUE && TrackIndex.Frames.back().PTS != AV_NOPTS_VALUE)
+        VP.Duration = (TrackIndex.Frames.back().PTS - TrackIndex.Frames.front().PTS) + LastFrameDuration;
+    
+    if (DurationHistogram.size() == 1 && MostCommonDuration.first > 0) {
+        // It's true CFR so make sure the frame rate matches the frame durations
+        av_reduce(&VP.FPS.Num, &VP.FPS.Den, VP.TimeBase.Den, MostCommonDuration.first * VP.TimeBase.Num, INT_MAX);
+    } else if (TrackIndex.Frames.size() >= 20 && DurationHistogram.size() > 1) {
+        // If the clip is long enough discard as many small duration bins as possible but less than 5% of the total number of frame durations and calculate a frame rate from that
+        size_t TotalHistogramFrames = TrackIndex.Frames.size() - 1;
+        size_t UsedHistogramFrames = TotalHistogramFrames - DurationHistogram[AV_NOPTS_VALUE];
+        DurationHistogram.erase(AV_NOPTS_VALUE);
+
+        while (DurationHistogram.size() > 1) {
+            const auto MinKey = std::min_element(DurationHistogram.begin(), DurationHistogram.end(), [](const std::pair<int64_t, size_t> &p1, const std::pair<int64_t, size_t> &p2) { return p1.second < p2.second; });
+            if (((UsedHistogramFrames - MinKey->second) * 100) / TotalHistogramFrames < 95)
+                break;
+            UsedHistogramFrames -= MinKey->second;
+            DurationHistogram.erase(MinKey);
+        }
+
+        if (!DurationHistogram.empty()) {
+            int64_t HistDuration = 0;
+            for (const auto &Iter : DurationHistogram)
+                HistDuration += Iter.first * Iter.second;
+
+            // FIXME, can this realistically overflow?
+            av_reduce(&VP.FPS.Num, &VP.FPS.Den, UsedHistogramFrames * VP.TimeBase.Den, HistDuration * VP.TimeBase.Num, INT_MAX);
+            NearestCommonFrameRate(VP.FPS);
+        }
+    } else if (VP.FPS.Num == 90000 && VP.FPS.Den == 1 && TrackIndex.Frames.size() >= 2) {
+        // This is the mpeg timebase and definitely not anywhere near the real fps so just fill in something more sane based on the duration of a single frame in the middle of the clip and hope it's good enough
+        // It's a fallback to make even obviously wrong mpeg timebase files have a saner framerate
+        int64_t F1 = TrackIndex.Frames[TrackIndex.Frames.size() / 2].PTS;
+        int64_t F2 = TrackIndex.Frames[TrackIndex.Frames.size() / 2 - 1].PTS;
+        if (F1 != AV_NOPTS_VALUE && F2 != AV_NOPTS_VALUE) {
+            av_reduce(&VP.FPS.Num, &VP.FPS.Den, VP.TimeBase.Den, (F1 - F2) * VP.TimeBase.Num, INT_MAX);
+            NearestCommonFrameRate(VP.FPS);
+        }
+    }
+
+    InitializeFormatSets();
+    SelectFormatSet(-1);
+
+    if (DefaultFormatSet.NumFrames != DefaultFormatSet.NumRFFFrames)
+        VP.FPS = OriginalFPS; // Restore the original FPS since it's generally always correct for files with RFF set
 
     Decoders[0] = std::move(Decoder);
 }
@@ -931,38 +1054,26 @@ void BestVideoSource::SetSeekPreRoll(int64_t Frames) {
 }
 
 bool BestVideoSource::IndexTrack(const ProgressFunction &Progress) {
-    std::unique_ptr<LWVideoDecoder> Decoder(new LWVideoDecoder(Source, HWDevice, ExtraHWFrames, VideoTrack, VariableFormat, Threads, LAVFOptions));
+    std::unique_ptr<LWVideoDecoder> Decoder(new LWVideoDecoder(Source, HWDevice, ExtraHWFrames, VideoTrack, ViewID, Threads, LAVFOptions));
 
     int64_t FileSize = Progress ? Decoder->GetSourceSize() : -1;
 
     TrackIndex.LastFrameDuration = 0;
-
-    // Fixme, implement frame discarding based on first seen format?
-    /*
-    bool First = true;
-    int Format = -1;
-    int Width = -1;
-    int Height = -1;
-    */
+    bool HasKeyFrames = false;
+    bool HasEarlyKeyFrames = false;
+    bool HasValidPTS = false;
 
     while (true) {
         AVFrame *F = Decoder->GetNextFrame();
         if (!F)
             break;
 
-        /*
-        if (First) {
-            Format = F->format;
-            Width = F->width;
-            Height = F->height;
-            First = false;
-        }
-        */
-
-        //if (VariableFormat || (Format == F->format && Width == F->width && Height == F->height)) {
-        TrackIndex.Frames.push_back({ F->pts, F->repeat_pict, !!(F->flags & AV_FRAME_FLAG_KEY), !!(F->flags & AV_FRAME_FLAG_TOP_FIELD_FIRST), GetHash(F) });
+        HasValidPTS = HasValidPTS || (F->pts != AV_NOPTS_VALUE);
+        HasKeyFrames = HasKeyFrames || !!(F->flags & AV_FRAME_FLAG_KEY);
+        if (TrackIndex.Frames.size() < 100)
+            HasEarlyKeyFrames = HasKeyFrames;
+        TrackIndex.Frames.push_back({ F->pts, F->repeat_pict, !!(F->flags & AV_FRAME_FLAG_KEY), !!(F->flags & AV_FRAME_FLAG_TOP_FIELD_FIRST), F->format, F->width, F->height, GetHash(F) });
         TrackIndex.LastFrameDuration = F->duration;
-        //}
 
         av_frame_free(&F);
         if (Progress) {
@@ -974,10 +1085,29 @@ bool BestVideoSource::IndexTrack(const ProgressFunction &Progress) {
     if (Progress)
         Progress(VideoTrack, INT64_MAX, INT64_MAX);
 
+    if (!TrackIndex.Frames.empty()) {
+        if (!HasKeyFrames) {
+            BSDebugPrint("No keyframes found when indexing which indicates an incorrectly flagged or very broken file, this may or may not cause performance problems when seeking");
+            for (auto &Iter : TrackIndex.Frames)
+                Iter.KeyFrame = true;
+        } else if (!HasEarlyKeyFrames) {
+            BSDebugPrint("No keyframes found in the first 100 frames when indexing, this may or may not cause performance problems when seeking");
+        }
+    }
+
+    if (!HasValidPTS) {
+        // Probably H264 in AVI
+        if (VP.Duration == TrackIndex.Frames.size()) {
+            // It's CFR so we know every frame duration is 1 unit
+            for (size_t i = 0; i < TrackIndex.Frames.size(); i++)
+                TrackIndex.Frames[i].PTS = i;
+        }
+    }
+
     return !TrackIndex.Frames.empty();
 }
 
-const VideoProperties &BestVideoSource::GetVideoProperties() const {
+const BSVideoProperties &BestVideoSource::GetVideoProperties() const {
     return VP;
 }
 
@@ -995,6 +1125,21 @@ BestVideoFrame *BestVideoSource::GetFrame(int64_t N, bool Linear) {
     if (N < 0 || N >= VP.NumFrames)
         return nullptr;
 
+    // Adjust frame number if an output format is chosen
+    if (VariableFormat >= 0 && FormatSets.size() > 1) {
+        const auto &ActiveSet = FormatSets[VariableFormat];
+        int64_t UsableFrames = 0;
+        int64_t SourceN = N;
+        for (const auto &Iter : TrackIndex.Frames) {
+            if (Iter.Format != ActiveSet.Format || Iter.Width != ActiveSet.Width || Iter.Height != ActiveSet.Height) {
+                N++;
+            } else {
+                if (UsableFrames++ == SourceN)
+                    break;
+            }
+        }
+    }
+
     std::unique_ptr<BestVideoFrame> F(FrameCache.GetFrame(N));
     if (!F)
         F.reset(Linear ? GetFrameLinearInternal(N) : GetFrameInternal(N));
@@ -1008,7 +1153,7 @@ void BestVideoSource::SetLinearMode() {
         BSDebugPrint("Linear mode is now forced");
         LinearMode = true;
         FrameCache.Clear();
-        for (size_t i = 0; i < MaxVideoSources; i++)
+        for (size_t i = 0; i < MaxVideoDecoders; i++)
             Decoders[i].reset();
     }
 }
@@ -1037,11 +1182,11 @@ namespace {
             Data.push_back(std::make_pair(F, GetHash(F)));
         }
 
-        size_t size() {
+        [[nodiscard]] size_t size() {
             return Data.size();
         }
 
-        size_t empty() {
+        [[nodiscard]] bool empty() {
             return Data.empty();
         }
 
@@ -1204,7 +1349,7 @@ BestVideoFrame *BestVideoSource::GetFrameInternal(int64_t N) {
         return GetFrameLinearInternal(N);
 
     // # 1 A suitable linear decoder exists and seeking is out of the question
-    for (int i = 0; i < MaxVideoSources; i++) {
+    for (int i = 0; i < MaxUsedVideoDecoders; i++) {
         if (Decoders[i] && Decoders[i]->GetFrameNumber() <= N && Decoders[i]->GetFrameNumber() >= SeekFrame)
             return GetFrameLinearInternal(N);
     }
@@ -1214,7 +1359,7 @@ BestVideoFrame *BestVideoSource::GetFrameInternal(int64_t N) {
     // Grab/create a new decoder to use for seeking, the position is irrelevant
     int EmptySlot = -1;
     int LeastRecentlyUsed = 0;
-    for (int i = 0; i < MaxVideoSources; i++) {
+    for (int i = 0; i < MaxUsedVideoDecoders; i++) {
         if (!Decoders[i])
             EmptySlot = i;
         if (Decoders[i] && DecoderLastUse[i] < DecoderLastUse[LeastRecentlyUsed])
@@ -1223,7 +1368,7 @@ BestVideoFrame *BestVideoSource::GetFrameInternal(int64_t N) {
 
     int Index = (EmptySlot >= 0) ? EmptySlot : LeastRecentlyUsed;
     if (!Decoders[Index])
-        Decoders[Index].reset(new LWVideoDecoder(Source, HWDevice, ExtraHWFrames, VideoTrack, VariableFormat, Threads, LAVFOptions));
+        Decoders[Index].reset(new LWVideoDecoder(Source, HWDevice, ExtraHWFrames, VideoTrack, ViewID, Threads, LAVFOptions));
 
     DecoderLastUse[Index] = DecoderSequenceNum++;
 
@@ -1236,7 +1381,7 @@ BestVideoFrame *BestVideoSource::GetFrameLinearInternal(int64_t N, int64_t SeekF
     int Index = -1;
     int EmptySlot = -1;
     int LeastRecentlyUsed = 0;
-    for (int i = 0; i < MaxVideoSources; i++) {
+    for (int i = 0; i < MaxUsedVideoDecoders; i++) {
         if (Decoders[i] && (!ForceUnseeked || !Decoders[i]->HasSeeked()) && Decoders[i]->GetFrameNumber() <= N && (Index < 0 || Decoders[Index]->GetFrameNumber() < Decoders[i]->GetFrameNumber()))
             Index = i;
         if (!Decoders[i])
@@ -1248,7 +1393,7 @@ BestVideoFrame *BestVideoSource::GetFrameLinearInternal(int64_t N, int64_t SeekF
     // If an empty slot exists simply spawn a new decoder there or reuse the least recently used decoder slot if no free ones exist
     if (Index < 0) {
         Index = (EmptySlot >= 0) ? EmptySlot : LeastRecentlyUsed;
-        Decoders[Index].reset(new LWVideoDecoder(Source, HWDevice, ExtraHWFrames, VideoTrack, VariableFormat, Threads, LAVFOptions));
+        Decoders[Index].reset(new LWVideoDecoder(Source, HWDevice, ExtraHWFrames, VideoTrack, ViewID, Threads, LAVFOptions));
     }
 
     std::unique_ptr<LWVideoDecoder> &Decoder = Decoders[Index];
@@ -1346,6 +1491,50 @@ bool BestVideoSource::InitializeRFF() {
     return true;
 }
 
+void BestVideoSource::InitializeFormatSets() {
+    std::map<std::tuple<int, int, int>, std::tuple<int64_t, int64_t, int64_t, bool>> SeenSets;
+    for (const auto &Iter : TrackIndex.Frames) {
+        auto V = std::make_tuple(Iter.Format, Iter.Width, Iter.Height);
+        if (SeenSets.insert(std::make_pair(V, std::make_tuple(0, 0, Iter.PTS, Iter.TFF))).second)
+            FormatSets.push_back(FormatSet{ {}, Iter.Format, Iter.Width, Iter.Height });
+        std::get<0>(SeenSets[V])++;
+        std::get<1>(SeenSets[V]) += Iter.RepeatPict + 2;
+    }
+
+    for (auto &Iter : FormatSets) {
+        auto V = std::make_tuple(Iter.Format, Iter.Width, Iter.Height);
+        Iter.NumFrames = std::get<0>(SeenSets[V]);
+        Iter.NumRFFFrames = std::get<1>(SeenSets[V]);
+        Iter.TFF = std::get<3>(SeenSets[V]);
+        if (std::get<2>(SeenSets[V]) != AV_NOPTS_VALUE)
+            Iter.StartTime = (static_cast<double>(VP.TimeBase.Num) * std::get<2>(SeenSets[V])) / VP.TimeBase.Den;      
+        Iter.VF.Set(av_pix_fmt_desc_get(static_cast<AVPixelFormat>(Iter.Format)));
+    }
+
+    DefaultFormatSet = FormatSets[0];
+    DefaultFormatSet.NumFrames = TrackIndex.Frames.size();
+    DefaultFormatSet.NumRFFFrames = 0;
+
+    for (auto &Iter : FormatSets) {
+        DefaultFormatSet.NumRFFFrames += Iter.NumRFFFrames;
+        Iter.NumRFFFrames = (Iter.NumRFFFrames + 1) / 2; // Can't round before adding it together
+
+        if (DefaultFormatSet.Format != Iter.Format)
+            DefaultFormatSet.Format = AV_PIX_FMT_NONE;
+        if (DefaultFormatSet.Width != Iter.Width || DefaultFormatSet.Height != Iter.Height) {
+            DefaultFormatSet.Width = 0;
+            DefaultFormatSet.Height = 0;
+        }
+    }
+
+    DefaultFormatSet.NumRFFFrames = (DefaultFormatSet.NumRFFFrames + 1) / 2;
+
+    if (DefaultFormatSet.Format != AV_PIX_FMT_NONE)
+        DefaultFormatSet.VF.Set(av_pix_fmt_desc_get(static_cast<AVPixelFormat>(DefaultFormatSet.Format)));
+    else
+        DefaultFormatSet.VF = {};
+}
+
 BestVideoFrame *BestVideoSource::GetFrameWithRFF(int64_t N, bool Linear) {
     if (RFFState == RFFStateEnum::Uninitialized)
         InitializeRFF();
@@ -1376,7 +1565,10 @@ BestVideoFrame *BestVideoSource::GetFrameWithRFF(int64_t N, bool Linear) {
 }
 
 BestVideoFrame *BestVideoSource::GetFrameByTime(double Time, bool Linear) {
-    int64_t PTS = static_cast<int64_t>(((Time * 1000 * VP.TimeBase.Den) / VP.TimeBase.Num) + .001);
+    if (!CanSeekByTime)
+        throw BestSourceException("Can't get frame by time, file has frames with unknown timestamps");
+
+    int64_t PTS = static_cast<int64_t>(((Time * VP.TimeBase.Den) / VP.TimeBase.Num) + .001);
     FrameInfo F{ PTS };
 
     auto Pos = std::lower_bound(TrackIndex.Frames.begin(), TrackIndex.Frames.end(), F, [](const FrameInfo &FI1, const FrameInfo &FI2) { return FI1.PTS < FI2.PTS; });
@@ -1389,28 +1581,64 @@ BestVideoFrame *BestVideoSource::GetFrameByTime(double Time, bool Linear) {
     return GetFrame(Frame - 1);
 }
 
+const std::vector<BestVideoSource::FormatSet> &BestVideoSource::GetFormatSets() const {
+    return FormatSets;
+}
+
+void BestVideoSource::SelectFormatSet(int Index) {
+    if (Index >= static_cast<int>(FormatSets.size()) || Index < -1)
+        throw BestSourceException("Invalid format set");
+    VariableFormat = Index;
+    BestVideoSource::FormatSet &SrcSet = (Index < 0) ? DefaultFormatSet : FormatSets[Index];
+
+    VP.VF = SrcSet.VF;
+    VP.Format = SrcSet.Format;
+    VP.Width = SrcSet.Width;
+    VP.Height = SrcSet.Height;
+
+    /* Height and width but largest possible size where it's a multiple of the subsampling */
+    VP.SSModWidth = VP.Width - (VP.Width % (1 << VP.VF.SubSamplingW));
+    VP.SSModHeight = VP.Height - (VP.Height % (1 << VP.VF.SubSamplingH));
+
+    VP.StartTime = SrcSet.StartTime;
+
+    VP.NumFrames = SrcSet.NumFrames;
+    VP.NumRFFFrames = SrcSet.NumRFFFrames;
+
+    VP.TFF = SrcSet.TFF;
+
+    RFFState = (VP.NumFrames == VP.NumRFFFrames) ? RFFStateEnum::Unused : RFFStateEnum::Uninitialized;
+}
+
+
 ////////////////////////////////////////
 // Index read/write
 
-typedef std::array<uint8_t, 13> VideoCompArray;
+typedef std::array<uint8_t, sizeof(int64_t) + sizeof(int) + sizeof(int) + sizeof(int) + sizeof(int) + sizeof(uint8_t)> VideoCompArray;
 
-static VideoCompArray GetVideoCompArray(int64_t PTS, int RepeatPict, bool KeyFrame, bool TFF) {
+static VideoCompArray GetVideoCompArray(int64_t PTS, int RepeatPict, int Format, int Width, int Height, bool KeyFrame, bool TFF) {
     VideoCompArray Result;
     memcpy(Result.data(), &PTS, sizeof(PTS));
     memcpy(Result.data() + sizeof(PTS), &RepeatPict, sizeof(RepeatPict));
+    memcpy(Result.data() + sizeof(PTS) + sizeof(RepeatPict), &Format, sizeof(Format));
+    memcpy(Result.data() + sizeof(PTS) + sizeof(RepeatPict) + sizeof(Format), &Width, sizeof(Width));
+    memcpy(Result.data() + sizeof(PTS) + sizeof(RepeatPict) + sizeof(Format) + sizeof(Width), &Height, sizeof(Height));
     uint8_t Flags = static_cast<uint8_t>(KeyFrame) | (static_cast<uint8_t>(TFF) << 1);
-    memcpy(Result.data() + sizeof(PTS) + sizeof(RepeatPict), &Flags, sizeof(Flags));
+    memcpy(Result.data() + sizeof(PTS) + sizeof(RepeatPict) + sizeof(Format) + sizeof(Width) + sizeof(Height), &Flags, sizeof(Flags));
+
+    static_assert(sizeof(PTS) + sizeof(RepeatPict) + sizeof(Format) + sizeof(Width) + sizeof(Height) + sizeof(Flags) == sizeof(VideoCompArray));
+
     return Result;
 }
 
-bool BestVideoSource::WriteVideoTrackIndex(const std::filesystem::path &CachePath) {
-    file_ptr_t F = OpenCacheFile(CachePath, Source, VideoTrack, true);
+bool BestVideoSource::WriteVideoTrackIndex(bool AbsolutePath, const std::filesystem::path &CachePath) {
+    file_ptr_t F = OpenCacheFile(AbsolutePath, CachePath, Source, VideoTrack, true);
     if (!F)
         return false;
     WriteBSHeader(F, true);
     WriteInt64(F, FileSize);
     WriteInt(F, VideoTrack);
-    WriteInt(F, VariableFormat);
+    WriteInt(F, ViewID);
     WriteString(F, HWDevice);
     WriteInt(F, ExtraHWFrames);
 
@@ -1438,7 +1666,7 @@ bool BestVideoSource::WriteVideoTrackIndex(const std::filesystem::path &CachePat
             LastPTSValue = OrigPTS;
         }
 
-        Dict.insert(std::make_pair(GetVideoCompArray(PTS, Iter.RepeatPict, Iter.KeyFrame, Iter.TFF), 0));
+        Dict.insert(std::make_pair(GetVideoCompArray(PTS, Iter.RepeatPict, Iter.Format, Iter.Width, Iter.Height, Iter.KeyFrame, Iter.TFF), 0));
     }
 
     // Only bother with a dictionary if it's not too big
@@ -1463,7 +1691,7 @@ bool BestVideoSource::WriteVideoTrackIndex(const std::filesystem::path &CachePat
                 LastPTSValue = OrigPTS;
             }
 
-            WriteByte(F, Dict[GetVideoCompArray(PTS, Iter.RepeatPict, Iter.KeyFrame, Iter.TFF)]);
+            WriteByte(F, Dict[GetVideoCompArray(PTS, Iter.RepeatPict, Iter.Format, Iter.Width, Iter.Height, Iter.KeyFrame, Iter.TFF)]);
             fwrite(Iter.Hash.data(), 1, Iter.Hash.size(), F.get());
         }
     } else {
@@ -1473,6 +1701,9 @@ bool BestVideoSource::WriteVideoTrackIndex(const std::filesystem::path &CachePat
             fwrite(Iter.Hash.data(), 1, Iter.Hash.size(), F.get());
             WriteInt64(F, Iter.PTS);
             WriteInt(F, Iter.RepeatPict);
+            WriteInt(F, Iter.Format);
+            WriteInt(F, Iter.Width);
+            WriteInt(F, Iter.Height);
             WriteByte(F, static_cast<uint8_t>(Iter.KeyFrame) | (static_cast<uint8_t>(Iter.TFF) << 1));
         }
     }
@@ -1480,8 +1711,8 @@ bool BestVideoSource::WriteVideoTrackIndex(const std::filesystem::path &CachePat
     return true;
 }
 
-bool BestVideoSource::ReadVideoTrackIndex(const std::filesystem::path &CachePath) {
-    file_ptr_t F = OpenCacheFile(CachePath, Source, VideoTrack, false);
+bool BestVideoSource::ReadVideoTrackIndex(bool AbsolutePath, const std::filesystem::path &CachePath) {
+    file_ptr_t F = OpenCacheFile(AbsolutePath, CachePath, Source, VideoTrack, false);
     if (!F)
         return false;
     if (!ReadBSHeader(F, true))
@@ -1490,7 +1721,7 @@ bool BestVideoSource::ReadVideoTrackIndex(const std::filesystem::path &CachePath
         return false;
     if (!ReadCompareInt(F, VideoTrack))
         return false;
-    if (!ReadCompareInt(F, VariableFormat))
+    if (!ReadCompareInt(F, ViewID))
         return false;
     if (!ReadCompareString(F, HWDevice))
         return false;
@@ -1518,6 +1749,9 @@ bool BestVideoSource::ReadVideoTrackIndex(const std::filesystem::path &CachePath
             FrameInfo FI = {};
             FI.PTS = ReadInt64(F);
             FI.RepeatPict = ReadInt(F);
+            FI.Format = ReadInt(F);
+            FI.Width = ReadInt(F);
+            FI.Height = ReadInt(F);
             uint8_t Flags = ReadByte(F);
             FI.KeyFrame = !!(Flags & 1);
             FI.TFF = !!(Flags & 2);
@@ -1541,6 +1775,9 @@ bool BestVideoSource::ReadVideoTrackIndex(const std::filesystem::path &CachePath
                 return false;
             FI.PTS = ReadInt64(F);
             FI.RepeatPict = ReadInt(F);
+            FI.Format = ReadInt(F);
+            FI.Width = ReadInt(F);
+            FI.Height = ReadInt(F);
             uint8_t Flags = ReadByte(F);
             FI.KeyFrame = !!(Flags & 1);
             FI.TFF = !!(Flags & 2);
@@ -1578,8 +1815,16 @@ void BestVideoSource::WriteTimecodes(const std::filesystem::path &TimecodeFile)
         throw BestSourceException("Couldn't open timecode file for writing");
 
     fprintf(F.get(), "# timecode format v2\n");
-    for (const auto &Iter : TrackIndex.Frames)
-        fprintf(F.get(), "%.02f\n", (Iter.PTS * VP.TimeBase.Num) / (double)VP.TimeBase.Den);
+    for (const auto &Iter : TrackIndex.Frames) {
+        double timestamp = ((Iter.PTS * VP.TimeBase.Num) / (double)VP.TimeBase.Den) * 1000;
+#ifdef __cpp_lib_to_chars
+        char buffer[100];
+        auto res = std::to_chars(buffer, buffer + sizeof(buffer), timestamp, std::chars_format::fixed, 2);
+        fprintf(F.get(), "%s\n", std::string(buffer, res.ptr - buffer).c_str());
+#else
+        fprintf(F.get(), "%.02f\n", timestamp);
+#endif
+    }
 }
 
 const BestVideoSource::FrameInfo &BestVideoSource::GetFrameInfo(int64_t N) const {
@@ -1588,4 +1833,14 @@ const BestVideoSource::FrameInfo &BestVideoSource::GetFrameInfo(int64_t N) const
 
 bool BestVideoSource::GetLinearDecodingState() const {
     return LinearMode;
-}
\ No newline at end of file
+}
+
+int BestVideoSource::SetMaxDecoderInstances(int NumInstances) {
+    if (NumInstances < 1 || NumInstances > MaxVideoDecoders)
+        MaxUsedVideoDecoders = MaxVideoDecoders;
+    else
+        MaxUsedVideoDecoders = NumInstances;
+    for (int i = NumInstances; i < MaxVideoDecoders; i++)
+        Decoders[i].reset();
+    return MaxUsedVideoDecoders;
+}
